# -*- coding: utf-8 -*-
import redis
import uuid
import urlparse
from contextlib import contextmanager

from .exceptions import MutexHeldError


def connect_redis(uri):
    """
    Create a redis connection by uri.
    """
    puri = urlparse.urlparse(uri)
    host = puri.hostname
    port = puri.port
    password = puri.password if puri.password else ''
    db_name = puri.path.split('/')[1]
    r = redis.Redis(host=host, port=port, password=password, db=db_name)
    assert r.ping()
    return r


@contextmanager
def mutex(client, name, expire):
    """
    :type client: redis.Redis
    """
    lock_id = str(uuid.uuid4())
    won = client.setnx(name, lock_id)
    if won:
        client.expire(name, expire)
        yield
        with client.pipeline(True) as pipe:
            try:
                pipe.watch(name)
                if pipe.get(name) == lock_id:
                    pipe.multi()
                    pipe.delete(name)
                    pipe.execute()
                pipe.unwatch()
            except redis.WatchError:
                pass
    else:
        raise MutexHeldError



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import datetime
import uuid
import collections
import unittest

from redisdict.redisdict import SimpleRedisDict, ComplexRedisDict
from redisdict.exceptions import SerialisationError


class _RedisDictTestCase(unittest.TestCase):
    klass = SimpleRedisDict
    name = 'dct'

    def setUp(self):
        self.klass(self.name, {}, autoclean=True)  # clean it up

    def test_init_empty(self):
        self.klass(self.name, {})

    def test_key_not_exist(self):
        cloud = self.klass(self.klass, {'name': 'Jim'})
        cloud['name']
        with self.assertRaises(KeyError):
            cloud['age']

    def test_delete(self):
        pass

    def test_resolve_options(self):
        pass


class SimpleRedisDictCase(_RedisDictTestCase):
    klass = SimpleRedisDict

    def test_dict(self):
        origin = {
            'name': 'Jim',
            'age': 5,
        }

        cloud = SimpleRedisDict(self.klass, origin)

        for k, v in origin.items():
            value_in_cloud = cloud[k]
            self.assertIsInstance(value_in_cloud, str)
            self.assertEqual(value_in_cloud, str(v))

    def test_raise_error(self):
        with self.assertRaises(SerialisationError):
            self.klass(self.name, {'name': None})

    def test_with_default_dict(self):
        v = 'value of default dict'
        dct = collections.defaultdict(lambda: v)
        cloud = self.klass(self.name, dct)
        self.assertEqual(cloud[str(uuid.uuid4())], v)


class ComplexRedisDictCase(_RedisDictTestCase):
    klass = ComplexRedisDict

    def test_dict(self):
        origin = {
            'name': 'Jim',
            'birth': datetime.date.today(),
            'id': uuid.uuid4(),
            'address': None,
        }

        cloud = self.klass(self.name, origin)

        for k, v in origin.items():
            self.assertEqual(v, cloud[k])



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import logging
logging.basicConfig()

VERSION = (0, 0, 2)
__version__ = '.'.join(map(str, VERSION))

logger = logging.getLogger('redisdict')

from .redisdict import RedisDict, SimpleRedisDict, ComplexRedisDict



#!/usr/bin/env python
# encoding: utf-8
from __future__ import absolute_import

import cPickle
from collections import Mapping, defaultdict


from . import logger
from .utils import connect_redis, mutex
from .exceptions import SerialisationError


class SimpleRedisDict(Mapping):
    def __init__(self, name, dct, uri='redis://127.0.0.1:6379/0', **kwargs):
        """
        :param name: Name to specific this object in Redis.
        :param dct: The dict will map to Redis.
        :type dct: dict
        :param uri: Redis uri.

        :param autoclean: If sets to False, will keep the existing key in redis hash, then mapping the others.
                          If sets to True, with delete the old redis hash first, then start mapping.
                          Default is **True** .

        """
        self.name = name

        self._name = self.generate_key_name(self.name)
        self._dct = dct.copy()
        self._conn = connect_redis(uri)
        self._default_key = self.generate_key_name('default')
        self._is_default_dict = False

        self.resolve_options(kwargs)

        for key, value in self._dct.items():
            self.setitem(key, value, force=False)

        if isinstance(self._dct, defaultdict):
            self._is_default_dict = True
            self.setitem(self._default_key, self._dct.default_factory())

        self.after_init()

    def after_init(self):
        pass

    def resolve_options(self, options):
        autoclean = options.pop('autoclean', True)
        if autoclean:
            self.clear()

    @classmethod
    def generate_key_name(cls, name):
        prefix = cls.__name__ + ':'
        return '{0}{1}'.format(prefix, name)

    @classmethod
    def dumps(cls, obj):
        if obj is None:
            raise SerialisationError('NoneType is not supported in {}'.format(cls.__name__))
        if not isinstance(obj, basestring):
            logger.warning("{0} => {1} will be converting to <type 'str'>".format(obj, type(obj)))
        return str(obj)

    @classmethod
    def loads(cls, b):
        return b

    def setitem(self, key, value, force=False):
        if (key not in self) or force:
            return bool(self._conn.hset(self._name, key, self.dumps(value)))
        return False

    def getitem(self, key):
        value = self._conn.hget(self._name, key=key)
        return self.loads(value)

    def clear(self):
        """
        Delete this object in Redis.
        """
        return self._conn.delete(self._name)

    def __getitem__(self, key):
        if (key not in self) and (not self._is_default_dict):
            raise KeyError(key)
        return self.getitem(key) or self.get(self._default_key)

    def __setitem__(self, key, value):
        return self.setitem(key, value, force=True)

    def __delitem__(self, key):
        return self._conn.hdel(self._name, key)

    def __len__(self):
        return self._conn.hlen(self._name)

    def __iter__(self):
        return (key for key in self._conn.hkeys(self._name) if key != self._default_key)

    def __contains__(self, item):
        return self._conn.hexists(self._name, item)

    def Lock(self, expire):
        """
        :type expire: int
        """
        name = '_LOCK_{0}'.format(self._name)
        return mutex(self._conn, name, expire)

    def __str__(self):
        return '<{0}>'.format(self._name)


class ComplexRedisDict(SimpleRedisDict):

    @classmethod
    def dumps(cls, obj):
        return cPickle.dumps(obj)

    @classmethod
    def loads(cls, b):
        return cPickle.loads(b)


RedisDict = SimpleRedisDict



import setuptools
import codecs


def get_requirements(filename):
    return codecs.open('requirements/' + filename, encoding='utf-8').read().splitlines()


setuptools.setup(
    name='redisdict',
    version='0.0.2',
    packages=setuptools.find_packages(exclude=['tests', 'tests.*']),
    url='https://github.com/kxrr/redisdict',
    license='MIT',
    author='kxrr',
    author_email='hi@kxrr.us',
    description='A dict-like object using Redis as the backend.',
    install_requires=get_requirements('default.txt'),
    tests_require=get_requirements('test.txt'),
)
